import pytest

from eerily.data.generators.events import PoissonEvents


@pytest.mark.parametrize(
    "rate, expected",
    [
        (0.5, [0, 1, 0, 0, 1, 0, 0, 0, 1, 1]),
        (0.1, [0, 0, 0, 0, 1, 0, 0, 0, 0, 0]),
        (0.9, [1, 1, 1, 1, 1, 0, 1, 1, 1, 1]),
    ],
)
def test_poisson_events(rate, expected):
    seed = 42
    pe = PoissonEvents(rate=rate, seed=seed)

    length = 10
    events = [next(pe) for _ in range(length)]

    assert events == expected



import copy
from dataclasses import dataclass
from typing import Dict, Iterator, List

import numpy as np

from eerily.data.generators.stepper import BaseStepper, StepperModelParams


@dataclass(frozen=True)
class ARModelParams(StepperModelParams):
    """Parameters of our AR model,

    $$s(t+1) = \phi_0 + \phi_1 s(t) + \epsilon.$$

    :param delta_t: step size of time in each iteration
    :param phi0: pho_0 in the AR model
    :param phi1: pho_1 in the AR model
    :param epsilon: noise iterator, e.g., Gaussian noise
    :param initial_state: a dictionary of the initial state, e.g., `np.array([1])`
    :param variable_name: variable names
    """

    delta_t: float
    phi0: float
    phi1: float
    epsilon: Iterator


class AR1Stepper(BaseStepper):
    """Stepper that calculates the next step in time in an AR model

    :param model_params: parameters for the AR model
    """

    def __next__(self):
        epsilon = next(self.model_params.epsilon)

        next_s = self.model_params.phi0 + self.model_params.phi1 * self.current_state + epsilon
        self.current_state = next_s

        return copy.deepcopy(self.current_state)



import numpy as np
import pytest

from eerily.data.generators.events import PoissonEvents
from eerily.data.generators.noises import LogNormalNoise
from eerily.data.generators.spiking import SpikingEventParams, SpikingEventStepper


@pytest.mark.parametrize(
    ("spike_rate," "spike_level_mu,spike_level_std," "background_mu,background_std," "expected"),
    [
        (
            0.5,
            1.7,
            0.05,
            1.5,
            0.1,
            [
                4.62035577,
                9.23559839,
                4.83095958,
                4.92368125,
                8.65248086,
                3.93449308,
                4.53935095,
                4.34217659,
                9.94351653,
                9.36061423,
            ],
        )
    ],
)
def test_spiking_event_stepper(spike_rate, spike_level_mu, spike_level_std, background_mu, background_std, expected):

    seed = 42

    spike = PoissonEvents(rate=spike_rate, seed=seed)
    spike_noise = LogNormalNoise(mu=spike_level_mu, std=spike_level_std, seed=seed)
    background = LogNormalNoise(mu=background_mu, std=background_std, seed=seed)

    sep = SpikingEventParams(
        initial_state=0,
        variable_names=["event"],
        spike=spike,
        spike_level=spike_noise,
        background=background,
    )

    se = SpikingEventStepper(model_params=sep)

    length = 10

    values = np.array([next(se) for _ in range(length)])
    expected = np.array(expected)

    np.testing.assert_allclose(values, expected)



import numpy as np
import pytest

from eerily.data.generators.ar import AR1Stepper, ARModelParams


@pytest.fixture
def ar_epsilon():
    class ZeroEpsilon:
        """Constant noise

        :param epsilon: the constant value to be returned
        """

        def __init__(self, epsilon=0):
            self.epsilon = epsilon

        def __next__(self):
            return self.epsilon

    return ZeroEpsilon(epsilon=0)


@pytest.fixture
def ar_params(ar_epsilon):

    model_params = ARModelParams(
        delta_t=0.1,
        phi0=0,
        phi1=1.1,
        epsilon=ar_epsilon,
        initial_state=np.array([-1]),
        variable_names=["v"],
    )

    return model_params


@pytest.fixture
def ar_stepper(ar_params):
    return AR1Stepper(model_params=ar_params)


def test_ar_stepper(ar_stepper, length):

    container = np.array([])
    for _ in range(length):
        container = np.append(container, next(ar_stepper))

    container_truth = np.array(
        [
            -1.1,
            -1.21,
            -1.331,
            -1.4641,
            -1.61051,
            -1.771561,
            -1.9487171,
            -2.14358881,
            -2.357947691,
            -2.5937424601,
        ]
    )

    np.testing.assert_allclose(container, container_truth)



from typing import Iterator


class Factory:
    """An experimental factor that generates a history based on a stepper and the given length."""

    def __init__(self):
        pass

    def __call__(self, stepper: Iterator, length: int):
        i = 0
        while i < length:
            yield next(stepper)
            i += 1



import torch
from torch.utils.data import Dataset


class SinTimeSeriesDataset(Dataset):
    def __init__(self, sequence_length, input_length, prediction_length, nodes) -> None:
        super().__init__()

        self.sequence_length = sequence_length
        self.prediction_length = prediction_length
        self.input_length = input_length

        assert self.sequence_length > self.prediction_length
        self.nodes = nodes

        self._gen_data()

    def _gen_data(self):

        series = []
        for i in range(self.nodes):
            series.append(torch.sin(torch.linspace(i, self.sequence_length + i, self.sequence_length + 1)))

        self.data = torch.stack(series)

    def _slice(self, index):
        return (
            self.data[:, index : index + self.input_length],
            self.data[
                :,
                index + self.input_length : index + self.input_length + self.prediction_length,
            ],
        )

    def __len__(self):
        return self.sequence_length - self.input_length

    def __getitem__(self, index):
        return self._slice(index)


if __name__ == "__main__":

    d = SinTimeSeriesDataset(100, 5, 1, 3)

    print(d.data.shape)

    print(d[0][0].shape, d[0][1].shape)
    print(d[0])



import copy
from typing import Dict, Iterator, Optional, Sequence, Union

from loguru import logger

from eerily.data.generators.stepper import BaseStepper


class ElasticityStepper(BaseStepper):
    """Generates the next time step for an given initial condition.

    We use the following formula to generate the data

    $$
    \ln Q' = \ln Q + \epsilon (\ln P' - \ln P)
    $$

    Define new log transformed variables to make this a linear relation

    $$
    y' = y + \epsilon (x' - x).
    $$

    For example, with initial condition

    ```
    initial_condition = {"price": 0, "sales": 10}
    ```

    For a deterministic model, we have

    ```python
    length = 10
    elasticity = [-3] * length
    prices = range(length)
    initial_condition = {"price: 0.5, "sale": 3}

    es = ElasticityStepper(
        initial_condition=initial_condition,
        elasticity=elasticity,
        prices=prices
    )

    next(es)
    ```

    !!! warning "Initial Condition"
        Initial condition is a dictionary with at least two keys `sale` and `price`.

        Note that the initial condition is NOT returned in the iterator.

    """

    def __init__(
        self,
        initial_condition: Dict[str, float],
        elasticity: Union[Iterator, Sequence],
        prices: Union[Iterator, Sequence],
        length: Optional[int] = None,
    ):
        self.initial_condition = copy.deepcopy(initial_condition)
        self.current_state = copy.deepcopy(initial_condition)
        self.elasticity = elasticity
        self.prices = prices

        self.length = length

        if not isinstance(self.prices, Iterator):
            self.length = len(self.prices)
            self.prices = iter(self.prices)

        if not isinstance(self.elasticity, Iterator):
            elasticity_length = len(self.elasticity)
            self.elasticity = iter(self.elasticity)
            if (self.length is not None) and (elasticity_length != self.length):
                logger.warning(
                    f"elasticity length {elasticity_length} is different "
                    f"from prices length {self.length}; "
                    "Setting length to the min of the two."
                )
            self.length = min(self.length, elasticity_length)

    def __iter__(self):
        return self

    def __next__(self):

        price = next(self.prices)
        elasticity = next(self.elasticity)
        sale = self.current_state["sale"] + elasticity * (price - self.current_state["price"])

        self.current_state["sale"] = sale
        self.current_state["price"] = price
        self.current_state["elasticity"] = elasticity

        return copy.deepcopy(self.current_state)

    def __repr__(self) -> str:
        return (
            "ElasticityStepper: \n"
            f"initial_condition: {self.initial_condition}\n"
            f"current_state: {self.current_state}"
        )



import numpy as np
import pytest

from eerily.data.generators.noises import LogNormalNoise


@pytest.mark.parametrize(
    "mu,std,expected",
    [
        (
            1,
            0.3,
            [
                2.978486018476435,
                1.9897415742703028,
                3.4046268976160845,
                3.604450799108567,
                1.5139005158889947,
                1.8392284222498556,
                2.824558646174046,
                2.4722464279201573,
                2.704615215288275,
                2.10451875643865,
            ],
        )
    ],
)
def test_lognormal_noise(mu, std, expected):
    seed = 42
    lnn = LogNormalNoise(mu=mu, std=std, seed=seed)

    length = 10
    values = np.array([next(lnn) for _ in range(length)])
    expected = np.array(expected)

    np.testing.assert_allclose(values, expected)



import copy
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List

import numpy as np


@dataclass(frozen=True)
class StepperModelParams:
    """Base Parameters for Stepper

    :param initial_state: the initial state, e.g., `np.array([1])`
    :param variable_name: variable names
    """

    initial_state: np.ndarray
    variable_names: List[str]


class BaseStepper(ABC):
    """A framework to evolve a DGP to the next step"""

    def __init__(self, model_params) -> None:
        self.model_params = model_params
        self.current_state = copy.deepcopy(self.model_params.initial_state)

    def __iter__(self):
        return self

    @abstractmethod
    def __next__(self):
        pass

    def get_iterator(self):
        return self.__iter__()

    def __str__(self) -> str:
        return f"Model Parameters: {self.model_params}"



import numpy as np
import pandas as pd
import pytest

from eerily.data.generators.brownian import BrownianMotionParams, BrownianMotionStepper
from eerily.data.generators.noises import GaussianNoise


@pytest.fixture
def gaussian_force(seed):
    return GaussianNoise(mu=0, std=1, seed=seed)


@pytest.fixture
def brownian_motion_params(gaussian_force):

    model_params = BrownianMotionParams(
        gamma=0,
        delta_t=0.1,
        force_densities=gaussian_force,
        initial_state=np.array([0]),
        variable_names=["v"],
    )

    return model_params


@pytest.fixture
def brownian_motion_stepper(brownian_motion_params):
    return BrownianMotionStepper(model_params=brownian_motion_params)


def test_brownian_motion_stepper(brownian_motion_stepper, length):

    container = np.array([])
    for _ in range(length):
        container = np.append(container, next(brownian_motion_stepper))

    container_truth = np.array(
        [
            0.030471707975443137,
            -0.07352670264860642,
            0.001518416932039315,
            0.09557488857116071,
            -0.09952863029422294,
            -0.2297465809804547,
            -0.21696254066372622,
            -0.24858679989808444,
            -0.25026691564851333,
            -0.3355713084058713,
        ]
    )

    np.testing.assert_allclose(container, container_truth)



import copy
from dataclasses import dataclass
from typing import Any, Dict, Iterator, List, Optional, Sequence, Union

import numpy as np

from eerily.data.generators.events import PoissonEvents
from eerily.data.generators.stepper import BaseStepper, StepperModelParams


@dataclass(frozen=True)
class SpikingEventParams(StepperModelParams):
    """
    Parameters for spiking events.

    :param spike: the spiking process, e.g., Poisson process.
    :param spike_level: the level of spikes, e.g. [`GaussianNoise`][eerily.data.generators.noise.GaussianNoise] with some positve mean. This parameter determines the height of the spikes.
    :param background: the stochastic noise level, e.g. [`GaussianNoise`][eerily.data.generators.noise.GaussianNoise].
    """

    spike: Iterator
    spike_level: Iterator
    background: Iterator


class SpikingEventStepper(BaseStepper):
    """Calculates the next step in a spiking event.

    :param model_params: a dataclass that contains the necessary parameters for the model.
        e.g., [`SpikingEventParams`][eerily.data.generators.spiking.SpikingEventParams]
    """

    def __next__(self) -> Dict[str, float]:

        background = next(self.model_params.background)
        spike = next(self.model_params.spike)
        spike_level = next(self.model_params.spike_level)

        v_next = background + spike * spike_level

        self.current_state = v_next

        return copy.deepcopy(self.current_state)



import numpy as np
import pytest

from eerily.data.generators.var import VAR1ModelParams, VAR1Stepper


@pytest.fixture
def var_epsilon():
    class ConstantEpsilon:
        """Constant noise

        :param epsilon: the constant value to be returned
        """

        def __init__(self, epsilon=0):
            self.epsilon = epsilon

        def __next__(self):
            return self.epsilon

    return ConstantEpsilon(epsilon=np.array([0, 0]))


@pytest.fixture
def var_params(var_epsilon):

    model_params = VAR1ModelParams(
        delta_t=0.01,
        phi0=np.array([0.1, 0.1]),
        phi1=np.array([[0.5, -0.25], [-0.35, 0.45 + 0.2]]),
        epsilon=var_epsilon,
        initial_state=np.array([1, 0]),
        variable_names=["s1", "s2"],
    )

    return model_params


@pytest.fixture
def var_stepper(var_params):
    return VAR1Stepper(model_params=var_params)


def test_var_stepper(var_stepper, length):

    container = []
    for _ in range(length):
        container.append(next(var_stepper).tolist())

    container = np.array(container)

    container_truth = np.array(
        [
            [6.00000000e-01, -2.50000000e-01],
            [4.62500000e-01, -2.72500000e-01],
            [3.99375000e-01, -2.39000000e-01],
            [3.59437500e-01, -1.95131250e-01],
            [3.28501562e-01, -1.52638437e-01],
            [3.02410391e-01, -1.14190531e-01],
            [2.79752828e-01, -8.00674820e-02],
            [2.59893285e-01, -4.99573532e-02],
            [2.42435981e-01, -2.34349292e-02],
            [2.27076723e-01, -8.52971532e-05],
        ]
    )

    np.testing.assert_allclose(container, container_truth)
